/*
  img2dsp.c

*/

#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>
#include <signal.h>
#include <strings.h>

#ifdef WIN32
#include <winsock.h>
#endif

#ifdef LINUX
#include <arpa/inet.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/select.h>
#include <netdb.h> 
#endif

#define CLIENT_APP
#include "../../firmware/ledmatrix.h"
#include "ir.h"
#include "image.h"

// input and output interfaces ...
static FILE *out = NULL;
static int sock = -1;

// signal handler to catch ctrl-c
void sigfunc(int sig) {
  if(sig != SIGINT) return;
  
  printf("Caught ctrl-c, exiting ...\n");

  if(sock >= 0) close(sock);

  exit(0);
}

void dsp_beep(int sock, struct sockaddr_in *addr) {
  unsigned char msg[] = { DSP_CMD_BEEP };

  if((sendto(sock, msg, sizeof(msg), 0, (struct sockaddr *)addr, 
	     sizeof(struct sockaddr_in))) != sizeof(msg))
    perror("sendto error on socket");
}

// mask the bits of a certain color (use to map background)
unsigned char get_mask(unsigned char byte) {
  unsigned char col = 0;  // color to use as mask
  unsigned char cmask, mask;
  int i;

  for(mask=0,cmask=0xc0;cmask;cmask >>= 2)
    if((byte & cmask) == (col & cmask)) mask |= cmask;

  return mask;
}

image_t *image_load(char *name) {

  if(strcasecmp(name+strlen(name)-4, ".rgb") == 0) 
    return rgb_read(name);
  else if(strcasecmp(name+strlen(name)-4, ".gif") == 0) 
    return gif_read(name);
    
  return NULL;
}

void image_free(image_t *img) {
  if(img) {
    free(img->data);
    free(img);
  }
}

int main(int argc, char **argv) {
  struct sockaddr_in dsp_addr, cli_addr;
  struct hostent *hp;
  int len, i;
  int is_c = 0;

  image_t *background = NULL;
  image_t *image = NULL;
  char tx_buffer[1+DSP_MEM];  // + 1 command byte

  int scroll = 0;
  
  opterr = 0;

  // parse command line options
  while ((i = getopt (argc, argv, "sb:")) != -1) {
    switch (i) {
      case 'b':
	printf("Loading background %s\n", optarg);
	if((background = image_load(optarg)) != NULL) {
	  if((background->width != DSP_W * DSP_PIX_X) ||
	     (background->height != DSP_H * DSP_PIX_Y)) {
	    printf("background image size is not %d x %d, ignoring\n",
		   DSP_W * DSP_PIX_X, DSP_H * DSP_PIX_Y);

	    image_free(background);
	    background = NULL;
	  }
	}
	break;

      case 's':
	scroll = 1;
	break;

      case '?':
	if (isprint (optopt))
	  fprintf (stderr, "Unknown option `-%c'.\n", optopt);
	else
	  fprintf (stderr,
		   "Unknown option character `\\x%x'.\n",
		   optopt);
	return 1;

      default:
	abort();
    }
  }

  // install signal handler if scroll option was given
  if(scroll) {
    printf("Enabling scroll, press ctrl-c to quit\n");
    signal(SIGINT, sigfunc);
  }

  if((argc-optind) != 2) {
    printf("Usage: img2dsp [options] infile.rgb outfile.bin\n");
    printf("    or img2dsp [options] infile.rgb outfile.c\n");
    printf("    or img2dsp [options] infile.rgb hostname\n");
    printf("Options:\n");
    printf("-s  scroll the image (only if hostname was given)\n");

#ifdef WIN32
    printf("Press return to continue ...\n");
    getchar();
#endif

    exit(1);
  }
 
  // load main image
  if((image = image_load(argv[optind])) == NULL) {
    exit(-1);
  }
  
  if((image->width < DSP_W * DSP_PIX_X) ||
     (image->height != DSP_H * DSP_PIX_Y)) {
    printf("image size is not %d x %d or wider!\n",
	   DSP_W * DSP_PIX_X, DSP_H * DSP_PIX_Y);
  }

  // try to resolv as a hostname
  if((hp = gethostbyname(argv[optind+1])) != NULL) {
    /* open datagram socket */
    if((sock = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
      perror("Could not create socket");
      exit(1);
    } 

    memset((char *) &dsp_addr, 0, sizeof(dsp_addr));
    dsp_addr.sin_family = AF_INET;
    dsp_addr.sin_addr.s_addr = *(int*)hp->h_addr;
    dsp_addr.sin_port = htons(DSP_PORT);

    cli_addr.sin_family = AF_INET;
    cli_addr.sin_addr.s_addr = htons(INADDR_ANY);
    cli_addr.sin_port = htons(DSP_PORT);
    
    if (bind(sock, (struct sockaddr *) &cli_addr, sizeof(cli_addr)) < 0) {
      perror("can't bind local address");
      exit(1);
    }

    printf("Sending via udp ...\n");
  }

  // hostname interpretation was not successful, use as filename
  if(sock < 0) {
    // file can only be written with exact size
    if((image->width != DSP_W * DSP_PIX_X) ||
       (image->height != DSP_H * DSP_PIX_Y)) {
      printf("image size is not %d x %d\n",
	     DSP_W * DSP_PIX_X, DSP_H * DSP_PIX_Y);
      exit(1);
    }

    /* check if target name ends with .c or .h */
    if((strcasecmp(argv[optind+1]+strlen(argv[optind+1])-2, ".c") == 0) ||
       (strcasecmp(argv[optind+1]+strlen(argv[optind+1])-2, ".h") == 0)) {
      printf("Converting to c source code ...\n");
      is_c = 1;
    } else { 
      printf("Converting to binary image ...\n");
    }
    
    /* open c file in ascii mode, otherwise in bin mode */
    if((out = fopen(argv[optind+1], is_c?"w":"wb")) == NULL) {
      perror("open(outfile)");
      exit(1);
    }

    // write c source header if approriate
    if(is_c) {
      char *name;

      if(strrchr(argv[optind+1], '/') != NULL)
	name = 1+strrchr(argv[optind+1], '/');
      else if(strrchr(argv[optind+1], '\\') != NULL)
	name = 1+strrchr(argv[optind+1], '\\');
      else
	name = argv[optind+1];

      fprintf(out, "/* %s */\n", name);
      
      // cut off .c or .h
      argv[optind+1][strlen(argv[optind+1])-2] = 0;
      fprintf(out, "unsigned char %s[] PROGMEM = {\n", name);
    }

    // write image
    for(i=0;i<DSP_MEM;i++) {
      if(is_c) 
	fprintf(out, "0x%02x,%s", image->data[i], ((i&15)==15)?"\n":" ");
      else
	fputc(image->data[i], out);
    }
    
    if(is_c) 
      fprintf(out, "};\n");
      
    fclose(out);

    // and exit as there's nothing to do in file mode
    exit(0);
  }

  // if we get here, we are in network mode
  do {
    int y, x;
    int fromlen, q, ack;
    unsigned char reply[256], *d;    
    fd_set rfds;
    struct timeval tv;
    
    if(background) {
      // copy mask to transmit buffer
      memcpy(tx_buffer+1, background->data, DSP_MEM);
      
      // and overlay image
      d = tx_buffer+1;
      for(y=0;y<DSP_H*DSP_PIX_Y;y++) {
	for(x=0;x<DSP_BPL;x++) {
	  unsigned char image_byte = image->data[y*image->width/4+x];
	  *d++ = 
	    (*d         &  get_mask(image_byte)) |
	    (image_byte & ~get_mask(image_byte));
	}
      }
    } else {
      // copy image to transmit buffer
      d = tx_buffer+1;
      for(y=0;y<DSP_H*DSP_PIX_Y;y++) 
	for(x=0;x<DSP_BPL;x++) 
	  *d++ = image->data[y*image->width/4+x];
    }
    
    ack = 0; 
    while(!ack) {
      
      tx_buffer[0] = DSP_CMD_IMAGE;
      if((sendto(sock, tx_buffer, DSP_MEM+1, 0, 
		 (struct sockaddr *)&dsp_addr, 
		 sizeof(dsp_addr))) != DSP_MEM+1)
	perror("sendto error on socket");
      
      FD_ZERO(&rfds);
      FD_SET(sock, &rfds);
      tv.tv_sec = 1;
      tv.tv_usec = 0;
      
      // wait 2 sec for ack
      if(( i = select(FD_SETSIZE, &rfds, NULL, NULL, &tv)) < 0) {
	perror("select");
	exit(1);
      }
      
      // timeout, no reply within 2 seconds
      if(i == 0) {
	// beep to notify user
	putchar('\007');
	fflush(stdout);
      }
      
      if(FD_ISSET(sock, &rfds)) {
	
	// wait for ack 
	fromlen = sizeof(dsp_addr);
	i = recvfrom(sock, reply, sizeof(reply),
		     0, (struct sockaddr *)&cli_addr, &fromlen);
	if (i < 0) { 
	  perror("recvfrom");  
	} else {
	  if((i == 2) && 
	     (reply[0] == DSP_CMD_ACK) &&
	     (reply[1] == DSP_CMD_IMAGE)) {
	    ack = 1;
	  } else if((i > 1) && (reply[0] == DSP_CMD_IR)) {
	    ir_receive(reply+1, i-1);
	  } else {
	    fprintf(stderr, "Unexpected reply message\n");
	  }
	}
      }
    }

    if(scroll) {
      // shift entire image left one pixel
      for(i=0;i<image->height;i++) {
	int j;
	unsigned char tmp, *data = image->data+i*(image->width/4);
	tmp = data[0];
	
	for(j=0;j<(image->width/4)-1;j++)
	  data[j] = (data[j] << 2) | (data[j+1] >> 6);
	
	data[j] = (data[j] << 2) | (tmp >> 6);	
      }
    }

  } while(scroll);
  
  close(sock);

  return 0;
}
